# -*- coding: utf-8 -*-
# !/usr/bin/env python
"""
    django-mongolog.  Simple Mongo based logger for Django
    Copyright (C) 2015 - John Furr

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from __future__ import print_function
import logging
from logging import Handler, NOTSET
from datetime import datetime as dt
import sys
try:
    from cStringIO import StringIO  # noqa
except ImportError:
    from io import StringIO  # noqa

import traceback
import json
import uuid
import pymongo
import requests
pymongo_version = int(pymongo.version.split(".")[0])
if pymongo_version >= 3:
    from pymongo.collection import ReturnDocument

from mongolog.models import LogRecord
from mongolog.exceptions import MissingConnectionError, LogConfigError

logger = logging.getLogger('')
console = logging.getLogger('mongolog-int')

uuid_namespace = uuid.UUID('8296424f-28b7-5982-a434-e6ec8ef529b3')


# TODO Move to mongolog.models.Mongolog
def get_mongolog_handler(logger_name=None, show_logger_names=False):
    """
    Return the first MongoLogHander found in the list of defined loggers.
    NOTE: If more than one is defined, only the first one is used.
    """
    if logger_name:
        logger_names = [logger_name]
    else:
        logger_names = [''] + list(logging.Logger.manager.loggerDict)

    if show_logger_names:
        console.info("get_mongolog_handler(): Logger_names: %s", json.dumps(logger_names, indent=4, sort_keys=True, default=str))

    for name in logger_names:
        logger = logging.getLogger(name)
        handler = None
        for _handler in logger.handlers:
            if isinstance(_handler, BaseMongoLogHandler):
                handler = _handler
                break
        if handler:
            console.debug("found handler: %s", handler)
            break

    if not handler:
        if logger_name:
            raise LogConfigError("logger '%s' does not have a mongolog based handler associated with it." % logger_name)

        raise LogConfigError("There are no loggers with a mongolog based handler.  Please see documentation about setting up LOGGING.")
    return handler


# Copied directly from python Formatter class
def formatException(ei):
    """
    Format and return the specified exception information as a string.

    This default implementation just uses
    traceback.print_exception()
    """
    sio = StringIO()

    traceback.print_exception(ei[0], ei[1], ei[2], None, sio)
    s = sio.getvalue()
    sio.close()
    if s[-1:] == "\n":
        s = s[:-1]
    return s


class BaseMongoLogHandler(Handler):

    REFERENCE = 'reference'
    EMBEDDED = 'embedded'

    def __init__(
            self, level=NOTSET, connection=None, database='mongolog', collection='mongolog', w=1, j=False, verbose=None,
            time_zone="local", record_type="embedded", max_keep=25, *args, **kwargs
        ):  # noqa
        super(BaseMongoLogHandler, self).__init__(level)
        self.connection = connection
        self.database = database
        self.collection = collection

        valid_record_types = [self.REFERENCE, self.EMBEDDED]
        if record_type not in valid_record_types:
            raise ValueError("record_type myst be one of %s" % valid_record_types)

        # The type of document we store
        self.record_type = record_type

        # number of dates to keep in embedded document
        self.max_keep = max_keep

        # The write concern
        self.w = w

        # If True block until write operations have been committed to the journal.
        self.j = j

        # Used to determine which time setting is used in the simple record_type
        self.time_zone = time_zone

        # If True will print each log_record to console before writing to mongo
        self.verbose = verbose

        if self.connection:
            self.connect()

            # Make sure the indexes are setup properly
            try:
                self.ensure_collections_indexed()
            except pymongo.errors.ServerSelectionTimeoutError:
                pass
        else:
            console.error("\n----------- Connection Error ------------")
            console.error("Hanlder(%s) missing 'connection' key", type(self))
            console.error("%s", json.dumps(self.__dict__, indent=4, sort_keys=True, default=str))
            console.error("------------------------------------------\n")
            raise MissingConnectionError("Missing 'connection' key")

    def __unicode__(self):
        return u'%s' % self.connection

    def __str__(self):
        return self.__unicode__()

    def connect(self, test=False):

        if pymongo_version == 3:
            self.client = self.connect_pymongo3(test)
        elif pymongo_version == 2:
            self.client = self.connect_pymongo2()

        # The mongolog database
        self.db = self.client[self.database]

        # This is the primary log document collection
        self.mongolog = self.db[self.collection]

        # This is the timestamp collection
        self.timestamp = self.db.timestamp

    def get_db(self):
        """
        Return a handler to the database handler
        """
        return getattr(self, "db", None)

    def get_timestamp_collection(self):
        return getattr(self, "timestamp", None)

    def get_collection(self):
        """
        Return the collection being used by MongoLogHandler
        """
        return getattr(self, "mongolog", None)

    def connect_pymongo3(self, test=False):

        try:
            if test:
                raise pymongo.errors.ServerSelectionTimeoutError("Just a test")

            if self.w == 0:
                # if w=0 you can't have other options
                self.client = pymongo.MongoClient(self.connection, serverSelectionTimeoutMS=5, w=self.w)
            else:
                self.client = pymongo.MongoClient(self.connection, serverSelectionTimeoutMS=5, w=self.w, j=self.j)

        except pymongo.errors.ServerSelectionTimeoutError:
            msg = "Unable to connect to mongo with (%s)" % self.connection
            # NOTE: Trying to log here ends up with Duplicate Key errors on upsert in emit()
            print(msg)
            raise pymongo.errors.ServerSelectionTimeoutError(msg)

        return self.client

    def connect_pymongo2(self):
        # TODO Determine proper try/except logic for pymongo 2.7 driver
        self.client = pymongo.MongoClient(self.connection)
        self.client.server_info()
        return self.client

    def new_key(self, old_key):
        """
        Repalce . and $ with Unicode full width equivalents
        """
        if "." in old_key:
            return old_key.replace(u".", u"．")
        elif "$" in old_key:
            return old_key.replace(u"$", u"＄")
        else:
            return old_key

    def check_keys(self, record):
        """
        Check for . and $ in two levels of keys below msg.
        TODO:   Make this a recursive function that looks for these keys
                n levels deep
        """
        if not isinstance(record['msg'], dict):
            return record

        for k, v in record['msg'].items():
            record['msg'][self.new_key(k)] = record['msg'].pop(k)

            if isinstance(v, dict):
                for old_key in record['msg'][k].keys():
                    record['msg'][k][self.new_key(old_key)] = record['msg'][k].pop(old_key)

        return record

    def create_log_record(self, record):
        """
        Convert the python LogRecord to a MongoLog Record.
        Also add a UUID which is a combination of the log message and log level.

        Override in subclasses to change log record formatting.
        See SimpleMongoLogHandler and VerboseMongoLogHandler
        """
        # This is still a python LogRecord Object that we are manipulating
        if record.exc_info:
            record.exc_text = formatException(record.exc_info)

        record = LogRecord(json.loads(json.dumps(record.__dict__, default=str)))
        if "mongolog.management.commands" in record['name']:
            return {'uuid': 'none', 'time': 'none', 'level': 'MONGOLOG-INTERNAL'}
        record = self.check_keys(record)

        # The UUID is a combination of the record.levelname and the record.msg
        if sys.version_info.major >= 3:
            uuid_key = str(record['msg']) + str(record['levelname'])
        else:
            uuid_key = (unicode(record['msg']) + unicode(record['levelname'])).encode('utf-8', 'replace')

        record.update({
            'uuid': uuid.uuid5(uuid_namespace, uuid_key).hex,
            # NOTE: if the user is using django and they have USE_TZ=True in their settings
            # then the timezone displayed will be what is specified in TIME_ZONE
            # For instance if they have TIME_ZONE='UTC' then both dt.now() and dt.utcnow()
            # will be equivalent.
            'time': dt.utcnow() if self.time_zone == 'utc' else dt.now(),
        })

        # If we are using an embedded document type
        # we need to create the dates array
        if self.record_type == self.EMBEDDED:
            record['dates'] = [record['time']]

        return record

    def ensure_collections_indexed(self):
        """
        Create the indexes if they are not already created
        """
        self.mongolog.create_index([("uuid", 1)], unique=True)
        self.mongolog.create_index([("dates", 1)])
        self.mongolog.create_index([("counter", 1)])

        self.timestamp.create_index([
            ("uuid", 1),
            ("ts", 1)
        ])

    def emit(self, record):
        """
        From python:  type(record) == LogRecord
        https://github.com/certik/python-2.7/blob/master/Lib/logging/__init__.py#L230
        """
        log_record = self.create_log_record(record)

        # TODO move this to a validate log_record method and add more validation
        log_record.get('uuid', ValueError("You must have a uuid in your LogRecord"))

        if self.verbose:
            print(json.dumps(log_record, sort_keys=True, indent=4, default=str))

        if self.record_type == self.EMBEDDED:
            self.insert_embedded(log_record)

        elif self.record_type == self.REFERENCE:
            if pymongo_version == 2:
                self.reference_log_pymongo_2(log_record)

            elif pymongo_version == 3:
                self.reference_log_pymongo_3(log_record)

    def insert_embedded(self, log_record):
        """
        Insert an embedded document.  Embedded documents have a 'counter'
        variable that increments each time the document is seen.  The 'date'
        array is capped at the last 'max_keep'
        """
        query = {'uuid': log_record['uuid']}
        result = self.mongolog.find(query)
        if result.count() == 0:

            # First time this record has been seen add a created field and insert it
            log_record['created'] = log_record.pop('time')
            log_record['counter'] = 1
            if pymongo_version == 2:
                self.mongolog.insert(log_record)
            elif pymongo_version == 3:
                self.mongolog.insert_one(log_record)
        else:
            # record has been seen before so we update the counter and push/pop
            # the log record time.  We keep the 'n' latest log record
            update = {
                "$push": {
                    'dates': {
                        '$each': [log_record['time']],
                        "$slice": -self.max_keep  # only keep the last n entries
                    }
                },
                # Keep a counter of the number of times we see this record
                "$inc": {'counter': 1}
            }

            if pymongo_version == 2:
                self.mongolog.update(query, update)
            elif pymongo_version == 3:
                self.mongolog.update_one(query, update)

    def reference_log_pymongo_2(self, log_record):
        query = {'uuid': log_record['uuid']}

        # remove the old document
        self.mongolog.find_and_modify(query, remove=True)

        # insert the new one
        self.mongolog.insert(log_record)

        # Add an entry in the timestamp collection
        self.timestamp.insert({
            'uuid': log_record['uuid'],
            'ts': log_record['time']
        })

    def reference_log_pymongo_3(self, log_record):
        query = {'uuid': log_record['uuid']}
        self.mongolog.find_one_and_replace(
            query,
            log_record,
            upsert=True,
            return_document=ReturnDocument.AFTER,
        )

        # Now update the timestamp collection
        # We can do this with a lower write concern than the previous operation since
        # we can alway's retreive the last datetime from the mongolog collection
        self.timestamp.insert({
            'uuid': log_record['uuid'],
            'ts': log_record['time']
        })


class SimpleMongoLogHandler(BaseMongoLogHandler):
    def create_log_record(self, record):
        record = super(SimpleMongoLogHandler, self).create_log_record(record)
        if record['uuid'] == 'none':
            return record
        mongolog_record = LogRecord({
            'name': record['name'],
            'thread': record['thread'],
            'process': record['process'],
            'level': record['levelname'],
            'msg': record['msg'],
            'path': record['pathname'],
            'module': record['module'],
            'line': record['lineno'],
            'func': record['funcName'],
            'filename': record['filename'],
            'uuid': record['uuid'],
            'time': record['time'],
        })

        if record.get('dates'):
            mongolog_record['dates'] = record['dates']

        # Add exception info
        if record['exc_info']:
            mongolog_record['exception'] = {
                'info': record['exc_info'],
                'trace': record['exc_text'].split("\n") if record['exc_text'] else None,
            }
        return mongolog_record


class VerboseMongoLogHandler(BaseMongoLogHandler):
    def create_log_record(self, record):
        record = super(VerboseMongoLogHandler, self).create_log_record(record)
        mongolog_record = LogRecord({
            'name': record['name'],
            'thread': {
                'num': record['thread'],
                'name': record['threadName'],
            },
            'process': {
                'num': record['process'],
                'name': record['processName'],
            },
            'level': {
                'name': record['levelname'],
                'num': record['levelno'],
            },
            'info': {
                'msg': record['msg'],
                'path': record['pathname'],
                'module': record['module'],
                'line': record['lineno'],
                'func': record['funcName'],
                'filename': record['filename'],
            },
            'uuid': record['uuid'],
            'time': record['time'],
        })

        if record['exc_info']:
            mongolog_record['exception'] = {
                'info': record['exc_info'],
                'trace': record['exc_text'],
            }

        return mongolog_record


class HttpLogHandler(SimpleMongoLogHandler):
    def __init__(self, level=NOTSET, client_auth='', timeout=3, verbose=False, time_zone="local", *args, **kwargs):
        # Make sure there is a trailing slash or reqests 2.8.1 will try a GET instead of POST
        self.client_auth = client_auth if client_auth.endswith('/') else "%s/" % client_auth

        self.timeout = timeout

        # Used to determine which time setting is used in the simple record_type
        self.time_zone = time_zone

        # Intentionally hard coded in HttpLogHandler
        self.record_type = 'reference'

        # If True will print each log_record to console before writing to mongo
        self.verbose = verbose

        # Don't call super here.  We don't want to call BaseMongoLogHandler.__init__ here.
        # But we still need this to be a python  Handler subclass with SimpleMongoLogger.create_log_record
        Handler.__init__(self, level=level)

    def __unicode__(self):
        return u'%s' % self.client_auth

    def emit(self, record):
        """
        From python:  type(record) == LogRecord
        https://github.com/certik/python-2.7/blob/master/Lib/logging/__init__.py#L230
        """
        log_record = self.create_log_record(record)

        # TODO move this to a validate log_record method and add more validation
        log_record.get('uuid', ValueError("You must have a uuid in your LogRecord"))
        if self.verbose:
            print("Inserting", json.dumps(log_record, sort_keys=True, indent=4, default=str))

        customer_id = self.client_auth.split("/")[-2]
        log_record['customer_id'] = customer_id

        r = requests.post(self.client_auth, json=json.dumps(log_record, default=str), timeout=self.timeout, proxies={'http':''})  # noqa
        # uncomment to debug
        try:
            print("Response:", json.dumps(r.json(), indent=4, sort_keys=True, default=str))
        except ValueError as e:
            if "No JSON object could be decoded" in str(e):
                print("log write failed: ", r)



#!/usr/bin/env python


class MissingConnectionError(ValueError):
    def __init__(self, *args, **kwargs):
        ValueError.__init__(self, *args, **kwargs)


class LogConfigError(ValueError):
    pass



#!/usr/bin/env python
"""
    django-mongolog.  Simple Mongo based logger for Django
    Copyright (C) 2015 - John Furr

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import pymongo
from pymongo import MongoClient
pymongo_version = int(pymongo.version.split(".")[0])
if pymongo_version >= 3:
    from pymongo.collection import ReturnDocument  # noqa: F401


class Mongolog(object):
    """
    This class provides a set of queriable functions.
    """
    # If LOGGER is None we wil try and find the first available logger
    LOGGER = None

    @classmethod
    def find(cls, logger=None, query=None, project=None, uuid=None, level=None, limit=None, **kwargs):
        """
         return self.collection.aggregate([
            {"$match": query},
            {"$project": proj},
            {"$sort": {'created': pymongo.DESCENDING}},
            {"$limit": limit},
        ])
        """
        from mongolog.handlers import get_mongolog_handler

        logger = cls.LOGGER if cls.LOGGER else logger

        handler = get_mongolog_handler(logger_name=logger)
        client = MongoClient(handler.connection)
        db = client.mongolog

        if logger:
            collection = getattr(db, handler.collection)
        else:
            collection = db.mongolog

        aggregate_commands = []

        if not query:
            query = {}

        if uuid:
            query.update({'uuid': uuid})

        if level:
            query.update({'level': level})

        if logger:
            query.update({'name': logger})

        aggregate_commands.append({"$match": query})

        if project:
            aggregate_commands.append({"$project": project})

        aggregate_commands.append({"$sort": {'created': pymongo.DESCENDING}})

        if limit:
            aggregate_commands.append({"$limit": limit})

        results = collection.aggregate(aggregate_commands)
        return results['result'] if isinstance(results, dict) else results


class LogRecord(dict):
    """
    {
        "_id" : ObjectId("565cc24656c02c294614f740"),
        "info" : {
            "module" : "<console>",
            "filename" : "<console>",
            "func" : "<module>",
            "msg" : "DEBUG message",
            "path" : "<console>",
            "line" : 1
        },
        "name" : "mongo",
        "thread" : {
            "num" : NumberLong(140061052233472),
            "name" : "MainThread"
        },
        "level" : {
            "num" : 10,
            "name" : "DEBUG"
        },
        "process" : {
            "num" : 10566,
            "name" : "MainProcess"
        },
        "time" : {
        "utc" : ISODate("2015-11-30T21:40:22.494Z"),
            "loc" : ISODate("2015-11-30T21:40:22.494Z")
        }
    }
    """
    pass



# -*- coding: utf-8 -*-
# !/usr/bin/env python
from __future__ import print_function
import logging
import logging.config
import json

import pymongo
pymongo_version = int(pymongo.version.split(".")[0])
if pymongo_version >= 3:
    from pymongo.collection import ReturnDocument  # noqa: F40

from mongolog.handlers import get_mongolog_handler

from django.core.management.base import BaseCommand

logger = logging.getLogger('console')


class Command(BaseCommand):

    def __init__(self, *args, **kwargs):
        super(Command, self).__init__(*args, **kwargs)
        self.prev_object_id = None

    def add_arguments(self, parser):
        parser.add_argument(
            '-l', '--limit', default=10, type=int, action='store', dest='limit',
            help='Limit Results',
        )
        parser.add_argument(
            '-t', '--tail', default=False, action='store_true', dest='tail',
            help='Tail the log file.  By default it will limit to 10 results.  Use --limit to change',
        )
        parser.add_argument(
            '-q', '--query', default=None, type=str, action='store', dest='query',
            help='Pass in a search query to mongo.',
        )

    def print_results(self, results):
        # older versions of pymongo didn't use a CommandCursor object to iterate over the results.
        # We check by trying to convert to a list and if there is a TypeError we continue on
        # and try to iterate over 'results' as thought it were a Cursor object.
        try:
            results = list(results['result'])
            results.reverse()
        except TypeError:
            pass

        for r in results:
            level = r.get('level', None)
            if level == 'INFO':
                logger.info(r)
            elif level == 'WARNING':
                logger.warn(r)
            elif level == 'ERROR':
                logger.error(r)
            elif level == 'DEBUG':
                logger.debug(r)
            elif level == 'CRITICAL':
                logger.critical(r)
            elif level == 'MONGOLOG-INTERNAL' or level is None:
                # Print nothing if it's a mongolog internal log
                pass
            else:
                raise Exception("level(%s) not found" % level)

    def fetch_results(self, options):
        query = options['query'] if options['query'] else {}
        proj = {'_id': 1, 'level': 1, 'msg': 1}
        limit = options['limit']
        return self.collection.aggregate([
            {"$match": query},
            {"$project": proj},
            {"$sort": {'created': pymongo.DESCENDING}},
            {"$limit": limit},
        ])

    def tail(self, options):
        initial = self.fetch_results(options)
        self.print_results(initial)
        raise NotImplementedError("--tail not finshed")

    def handle(self, *args, **options):
        if options['query']:
            options['query'] = json.loads(options['query'])

        handler = get_mongolog_handler('simple')
        self.collection = handler.get_collection()

        if options['tail']:
            self.tail(options)
        else:
            results = self.fetch_results(options)
            self.print_results(results)



import logging
logger = logging.getLogger('')

from django.views.generic.base import TemplateView


class HomeView(TemplateView):
    template_name = "mongolog/home.html"

#    def get_context_data(self, **kwargs):
#        context = super(HomeView, self).get_context_data(**kwargs)
#
#        try:
#            raise ValueError("ERROR MESSAGE2")
#        except Exception:
#            logger.exception({
#                'HomeView Test': True,
#                'stack trace?': "pleace?"
#            })
#            raise
#
#        return context



# -*- coding: utf-8 -*-
"""
    django-mongolog.  Simple Mongo based logger for Django
    Copyright (C) 2015 - John Furr

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import unittest
import logging
from logging import config  # noqa
import os
import sys
import subprocess
import time
import json
from unittest import skipIf
from requests.exceptions import ConnectionError

# Different imports for python2/3
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO  # noqa: F401

import pymongo
pymongo_major_version = int(pymongo.version.split(".")[0])

from mongolog.handlers import (
    get_mongolog_handler, SimpleMongoLogHandler
)
from mongolog.models import Mongolog

import django
django_version = django.VERSION[0]

from django.core.management import call_command
from django.test import TestCase
from django.test import Client

if django_version < 2:
    from django.core.urlresolvers import reverse
else:
    from django.urls import reverse

from django.conf import settings
LOGGING = settings.LOGGING
console = logging.getLogger("console")

console.info(
    "Available LOGGERS(%s)" % json.dumps(
        logging.Logger.manager.loggerDict, indent=4, sort_keys=True, default=str))


"""
    All log tests should log a dictionary with a 'test' key
    logger.info({
        'test': True,
        'msg': 'special message'
    })

    This key will allow us to easily clean test entries out of the
    database.

    NOTE: You can add any other key you want for testing purposes
"""
TEST_MSG = {
    'test': True,  # String so we can remove test entries
    'test class': 'TestBaseMongoLogHandler',
    'Life': {
        'Domain': {
            'Bacteria': [
                {
                    'name': ValueError,
                    'description': 'Just a bad description',
                    'unicode_test': 'ऄࢩ ࢩ ࢩǢ',
                }
            ],
            'Archaea': [],
            'Eukaryota': [
                {
                    'name': 'Excavata',
                    'description': 'Various flagellate protozoa',
                },
                {
                    'name': 'Amoebozoa',
                    'descritpion': 'most lobose amoeboids and slime moulds',
                },
                {
                    'name': 'Opisthokonta',
                    'description': 'animals, fungi, choanoflagellates, etc.',
                },
                {
                    'name': 'Rhizaria',
                    'description': 'Foraminifera, Radiolaria, and various other amoeboid protozoa'
                },
                {
                    'name': 'Chromalveolata',
                    'description': 'Stramenopiles (Brown Algae, Diatoms etc.)'
                },
                {
                    'name': 'Archaeplastida',
                    'description': 'Land plants, green algae, red algae, and glaucophytes'
                },
            ]
        }
    }
}


def raiseException(logger=None):
    """
    Used to test logger.exceptions.
    Use like:
    with self.assertRaises(ValueError):
        raiseException()
    """
    try:
        raise ValueError("Test Error")
    except ValueError:
        logger.exception(TEST_MSG)
        raise


class TestRemoveEntriesMixin(object):
    def remove_test_entries(self, test_key="msg.test"):
        """
        Remove all current test entries.
        Called in setUp and tearDown.
        TODO: Remove entries from timestamp collection
        """
        # Drop the timetamp collection completely
        self.handler.get_db().command("dropDatabase")

        # Ensure that we don't have any test entries
        self.assertEqual(0, self.collection.find({test_key: True}).count())


class TestBaseMongoLogHandler(TestCase, TestRemoveEntriesMixin):
    def setUp(self):
        self.logger = logging.getLogger("test.base.reference")

        self.handler = get_mongolog_handler("test.base.reference", show_logger_names=True)
        console.error("self.handler(%s)" % self.handler)
        self.collection = self.handler.get_collection()

        self.remove_test_entries()

    def test_middleware(self):
        console.debug(self)
        c = Client()
        response = c.get(reverse("home"))
        self.assertContains(response, "HERE YOU ARE ON monglog/home.html")

    def test_dot_in_key(self):
        console.debug(self)
        self.logger.info({
            'META': {
                'user.name': 'jfurr',
                'user$name': 'jfurr'
            },
            'user.name': 'jfurr',
            'user$name': 'jfurr'
        })

    def test_write_concern(self):
        console.debug(self)
        self.logging = logging.getLogger("test.base.reference.w0")
        self.test_basehandler_exception()

    def test_valid_record_type(self):
        console.debug(self)
        record_type = LOGGING['handlers']['test_base_invalid']['record_type']
        LOGGING['handlers']['test_base_invalid']['record_type'] = 'invalid type'
        with self.assertRaises(ValueError):
            logging.config.dictConfig(LOGGING)

        # Reset the log handler
        LOGGING['handlers']['test_base_invalid']['record_type'] = record_type
        logging.config.dictConfig(LOGGING)

    def test_connection_error(self):
        console.debug(self)
        if pymongo_major_version >= 3:
            with self.assertRaises(pymongo.errors.ServerSelectionTimeoutError):
                self.handler.connect(test=True)

    def test_basehandler_exception(self):
        console.debug(self)
        with self.assertRaises(ValueError):
            raiseException(self.logger)

        records = self.collection.find({
            'msg.test': True,
            'msg.Life.Domain.Eukaryota.name': "Archaeplastida",
            'levelname': 'ERROR'
        })
        self.assertEqual(1, records.count())

        expected_keys = [
            u'threadName',
            u'name',
            u'thread',
            u'relativeCreated',
            u'process',
            u'args',
            u'filename',
            u'module',
            u'funcName',
            u'levelno',
            u'processName',
            u'created',
            u'msecs',
            u'msg',
            u'exc_info',
            u'exc_text',
            u'pathname',
            u'_id',
            u'levelname',
            u'lineno',
            u'time',
            u'uuid',
        ]
        # To make test pass on python 3 version
        if sys.version_info[0] >= 3:
            expected_keys.append(u'stack_info')

        record = records[0]
        self.assertEqual(
            set(record.keys()),
            set(expected_keys)
        )

        self.assertIn('tests.py', record['pathname'])

        # Now test that the nested ValueError was successfully converted to a unicode str.
        try:
            # unicode will throw a NameError in Python 3
            self.assertEqual(unicode, type(record['msg']['Life']['Domain']['Bacteria'][0]['name']))
        except NameError:
            self.assertEqual(str, type(record['msg']['Life']['Domain']['Bacteria'][0]['name']))

        self.logger.info("Just some friendly info")
        self.logger.error("Just some friendly info")
        self.logger.debug("Just some friendly info")

    def test_str_unicode_mongologhandler(self):
        console.debug(self)
        self.assertEqual(self.handler.connection, u"%s" % self.handler)
        self.assertEqual(self.handler.connection, "%s" % self.handler)


class TestSimpleMongoLogHandler_Embedded(unittest.TestCase, TestRemoveEntriesMixin):
    def setUp(self):
        console.debug(self)
        self.logger = logging.getLogger('test.embedded')
        self.handler = get_mongolog_handler('test.embedded')
        self.collection = self.handler.get_collection()

        self.remove_test_entries()

        self.expected_keys = set([
            'dates',
            'uuid',
            'thread',
            'level',
            'process',
            'exception',
            'module',
            'filename',
            'func',
            'created',
            'msg',
            'path',
            'line',
            '_id',
            'name',
            'counter',
        ])

    def test_missing_connection_key(self):
        LOGGING['handlers']['simple_no_connection'] = {
            'level': 'DEBUG',
            # Uncomment section to play with SimpleMongoLogHandler
            'class': 'mongolog.SimpleMongoLogHandler',
        }
        LOGGING['loggers']['simple.no.connection'] = {
            'level': 'DEBUG',
            'handlers': ['simple_no_connection'],
            'propagate': True
        }
        with self.assertRaises(ValueError):
            logging.config.dictConfig(LOGGING)

        del LOGGING['handlers']['simple_no_connection']
        del LOGGING['loggers']['simple.no.connection']

    def test_exception(self):
        console.debug(self)
        with self.assertRaises(ValueError):
            raiseException(self.logger)

        records = self.collection.find({
            'msg.test': True,
            'msg.Life.Domain.Eukaryota.name': "Archaeplastida",
            'level': 'ERROR'
        })
        self.assertEqual(1, records.count())

        record = records[0]
        self.assertEqual(
            set(record.keys()),
            self.expected_keys
        )

        # Verify that the dates field has 1 item
        self.assertEqual(1, len(record['dates']))

        # now create anthe rduplicate entry and show that only one record is still present
        # however now it should have 2 values in 'dates'
        with self.assertRaises(ValueError):
            raiseException(self.logger)

        records = self.collection.find({
            'msg.test': True,
            'msg.Life.Domain.Eukaryota.name': "Archaeplastida",
            'level': 'ERROR'
        })
        self.assertEqual(1, records.count())
        record = records[0]
        # should have two values in dates now
        self.assertEqual(2, len(record['dates']))

        # Now check the exceptoin keys
        self.assertEqual(
            set(record['exception'].keys()),
            set(['info', 'trace'])
        )


class TestSimpleMongoLogHandler_Reference(unittest.TestCase, TestRemoveEntriesMixin):
    """
    Test the mongolog reference storage capability.
    In 'reference' mode there are two collections created
    1) mongolog that holds that latest record and a single 'date'
    2) timestamp collection that holds that individual timestamps

    These two collections are related to each other via the uuid key.
    """
    def setUp(self):
        self.logger = logging.getLogger('test.reference')
        self.handler = get_mongolog_handler('test.reference')
        self.collection = self.handler.get_collection()
        self.remove_test_entries()

    def test_logstructure_simple_reference(self):
        """
        Test the simple log record structure
        """
        console.debug(self)

        # Used in test dictionaries to test mongo serialization
        SMH_Obj = SimpleMongoLogHandler(connection=LOGGING['handlers']['simple']['connection'])

        # now test a serializable dict with an exception call
        log_msg = {'test': True, 'fruits': ['apple', 'orange'], 'error': str(ValueError), 'handler': str(SMH_Obj)}
        expected_keys = set([
            '_id', 'exception', 'name', 'thread', 'time',
            'process', 'level', 'msg', 'path', 'module',
            'line', 'func', 'filename', 'uuid'
        ])

        try:
            raise ValueError
        except ValueError:
            self.logger.exception(log_msg)

        rec = self.collection.find_one({'msg.fruits': ['apple', 'orange']})
        self.assertEqual(set(rec.keys()), expected_keys)

        # Python 2 duplicate entry test
        log_msg = {'test': True, 'fruits': ['apple', 'orange'], 'error': str(ValueError), 'handler': str(SMH_Obj)}
        try:
            raise ValueError
        except ValueError:
            self.logger.exception(log_msg)

        rec = self.collection.find_one({'msg.fruits': ['apple', 'orange']})
        self.assertEqual(set(rec.keys()), expected_keys)

        # Now try an exception log with a complex log msg.
        try:
            raise ValueError
        except ValueError:
            log_msg = {
                'test': True,
                'fruits': [
                    'apple',
                    'orange',
                    {'tomatoes': ['roma', 'kmato', 'cherry', ValueError, 'plum']},
                    {},
                    {}
                ],
                'object': SimpleMongoLogHandler,
                'instance': SMH_Obj,
            }
            self.logger.exception(log_msg)

        rec = self.collection.find_one({'msg.fruits': {'$in': ['apple', 'orange']}})
        self.assertEqual(set(rec.keys()), expected_keys)


class TestVerboseMongoLogHandler(unittest.TestCase, TestRemoveEntriesMixin):
    def setUp(self):
        console.debug(self)
        self.logger = logging.getLogger("test.verbose")
        self.handler = get_mongolog_handler("test.verbose")
        self.collection = self.handler.get_collection()

        self.remove_test_entries(test_key="info.msg.test")

    def test_logstructure_verbose_exception(self):
        """
        Test the verbose log record strucute
        """
        console.debug(self)
        with self.assertRaises(ValueError):
            raiseException(self.logger)

        records = self.collection.find({
            'info.msg.test': True,
            'info.msg.Life.Domain.Eukaryota.name': "Archaeplastida",
            'level.name': 'ERROR'
        })
        self.assertEqual(1, records.count())

        record = records[0]
        self.assertEqual(
            set(record.keys()),
            set(['info', 'exception', 'time', 'level', 'name', 'thread', 'process', '_id', 'uuid'])
        )

        self.assertEqual(
            set(record['info'].keys()),
            set(["filename", "func", "line", "module", "msg", "path"])
        )

        for entry in ['thread', 'process', 'level']:
            self.assertEqual(
                set(record[entry].keys()),
                set(['name', 'num'])
            )

        self.assertIn('tests.py', record['info']['path'])

    def test_logstructure_verbose_debug_info_warn(self):
        """
        TODO:  Beef this test up a bit
        """
        console.debug(self)
        self.handler.setLevel("WARNING")
        log_msg = {'test': True, 'msg': 'WARNING', 'msg2': 'DANGER'}
        query = {
            'info.msg.test': log_msg['test'],
            'info.msg.msg': log_msg['msg'],
            'info.msg.msg2': log_msg['msg2'],
            'level.name': 'WARNING'
        }
        self.logger.warn(log_msg)
        self.assertEqual(1, self.collection.find(query).count())

        rec = self.collection.find_one(query)
        self.assertEqual(
            set(rec.keys()),
            set(['info', 'name', 'thread', 'level', 'process', 'time', '_id', 'uuid'])
        )

        for key in ['process', 'level', 'thread']:
            self.assertEqual(
                set(rec[key].keys()),
                set(['num', 'name'])
            )

        self.assertEqual(rec['thread']['name'], "MainThread")
        self.assertEqual(rec['info']['filename'], "tests.py")
        self.assertEqual(rec['process']['name'], "MainProcess")


class TestHttpLogHandler(unittest.TestCase):
    def setUp(self):
        console.debug(self)
        self.logger = logging.getLogger("test.http")
        self.handler = get_mongolog_handler("test.http")
        self.collection = self.handler.get_collection()
        console.error("self.collection(%s)" % self.collection)

    @skipIf(sys.version_info.major == 3, "SKipping TestHttpLogHandler.test_timeout because of python version")
    def test_timeout(self):
        console.debug(self)
        timeout = LOGGING['handlers']['test_http_invalid']['timeout']
        LOGGING['handlers']['test_http_invalid']['timeout'] = 1

        logging.config.dictConfig(LOGGING)

        self.logger = logging.getLogger("test.http")

        with self.assertRaises(ConnectionError):
            self.logger.warn("Danger Will Robinson!")

        LOGGING['handlers']['test_http_invalid']['timeout'] = timeout
        logging.config.dictConfig(LOGGING)

    @skipIf(sys.version_info.major == 3, "SKipping TestHttpLogHandler.test_invalid_connection because of python version")
    def test_invalid_connection(self):
        console.debug(self)
        with self.assertRaises(ConnectionError):
            self.logger.warn("Danger Will Robinson!")


class TestManagementCommands(unittest.TestCase, TestRemoveEntriesMixin):
    def setUp(self):
        console.debug(self)
        self.logger = logging.getLogger('test.reference')
        self.handler = get_mongolog_handler('test.reference')
        self.collection = self.handler.get_collection()

        self.remove_test_entries()
        self.remove_test_entries(test_key='info.msg.test')

    def test_analog(self):
        console.debug(self)
        self.logger.debug({'test': True, 'logger': 'Debug'})
        self.logger.info({'test': True, 'logger': 'Info'})
        self.logger.warn({'test': True, 'logger': 'Warn'})
        self.logger.error({'test': True, 'logger': 'Error'})
        self.logger.critical({'test': True, 'logger': 'Critical'})

        call_command('analog', limit=10, query='{"name": "root"}')

        with self.assertRaises(NotImplementedError):
            call_command('analog', limit=20, tail=True)


class TestPerformanceTests(unittest.TestCase, TestRemoveEntriesMixin):
    def setUp(self):
        self.handler = get_mongolog_handler('test.embedded')

        # Get some pointers to the collections for easy querying
        self.collection = self.handler.get_collection()
        self.timestamp = self.handler.get_timestamp_collection()

        self.remove_test_entries()

        # TODO Do you need this?
        self.remove_test_entries(test_key='msg.Test')
        self.iterations = 100

    def _check_embedded_results(self, results, iterations):
        self.assertEqual(1, results.count())
        rec = results[0]
        self.assertEqual(iterations, rec['counter'])

        expected_date_len = iterations
        if iterations > self.handler.max_keep:
            expected_date_len = self.handler.max_keep

        if len(rec['dates']) > expected_date_len:
            console.error("GREATER: date(%s)", rec['dates'])

        self.assertEqual(len(rec['dates']), expected_date_len)

    def test_embedded(self):
        console.debug(self)
        self.logger = logging.getLogger('test.embedded')
        console.info("Starting embedded test:  max_keep(%s) iteration(%s)", self.handler.max_keep, self.iterations)

        start = time.time()
        for i in range(self.iterations):
            self.logger.info({'Test': True, 'Test1': 1})
            results = self.collection.find({'msg.Test': True, 'msg.Test1': 1})
            self._check_embedded_results(results, i + 1)

        end = time.time()
        results = self.collection.find({'msg.Test': True})
        self._check_embedded_results(results, self.iterations)
        console.warn("Test time: %s", end - start)

        # rerun with larger max_keep
        max_keep = LOGGING['handlers']['test_embedded']['max_keep']
        LOGGING['handlers']['test_embedded']['max_keep'] = 50
        logging.config.dictConfig(LOGGING)
        self.setUp()
        self.logger = logging.getLogger('test.embedded')

        console.info("Starting embedded test:  max_keep(%s) iteration(%s)", self.handler.max_keep, self.iterations)

        start = time.time()
        for i in range(self.iterations):
            self.logger.info({'Test': True})
            results = self.collection.find({'msg.Test': True})
            self._check_embedded_results(results, i + 1)
        end = time.time()
        results = self.collection.find({'msg.Test': True})
        self._check_embedded_results(results, self.iterations)
        console.warn("Test time: %s", end - start)

        LOGGING['handlers']['test_embedded']['max_keep'] = max_keep
        logging.config.dictConfig(LOGGING)
        self.setUp()
        self.logger = logging.getLogger('test.embedded')

    def _check_reference_results(self, results, i):
        # confirm that only one document is ever created for the same message
        self.assertEqual(1, len(results))
        uuid = results[0]['uuid']
        ts_results = self.timestamp.find({'uuid': uuid})
        td_results = list(ts_results)
        self.assertEqual(i + 1, len(td_results))

    def test_reference(self):
        console.debug(self)
        self.logger = logging.getLogger('test.reference')
        console.info("Starting embedded test:  max_keep(%s) iteration(%s)", self.handler.max_keep, self.iterations)

        start = time.time()
        for i in range(self.iterations):
            self.logger.info({'Test': True, 'Test1': 1})
            results = list(self.collection.find({'msg.Test': True, 'msg.Test1': 1}))

            self._check_reference_results(results, i)

        end = time.time()
        self.logger.info("Total time %s" % (end - start))


class MongoLogUtilsTests(unittest.TestCase, TestRemoveEntriesMixin):
    def setUp(self):
        self.handler = get_mongolog_handler('test.embedded')

        # Get some pointers to the collections for easy querying
        self.collection = self.handler.get_collection()
        self.timestamp = self.handler.get_timestamp_collection()
        self.logger = logging.getLogger('test.embedded')

        self.remove_test_entries()

    def test_find_for_embedded(self):
        console.debug(self)
        Mongolog.LOGGER = 'test.embedded'
        info_items = [
            {'location': {'country': 'USA', 'state': 'WA', 'city': 'Puyallup'}},
            {'location': {'country': 'USA', 'state': 'WA', 'city': 'Seattle'}},
            {'location': {'country': 'USA', 'state': 'MA', 'city': 'Boston'}},
            {'location': {'country': 'USA', 'state': 'MS', 'city': 'Greenville'}},
            {'location': {'country': 'France', 'region': 'Normandy', 'city': 'Rouen'}},
            {'location': {'country': 'France', 'region': 'Île-de-France', 'city': 'Paris'}},
        ]
        for item in info_items:
            self.logger.info(item)

        warn_items = [
            {'location': {'country': 'France', 'region': 'Normandy', 'city': 'Caen'}},
            {'location': {'country': 'France', 'region': 'Normandy', 'city': 'Bayeux'}},
            {'location': {'country': 'USA', 'state': 'MA', 'city': 'Framingham'}},
        ]
        for item in warn_items:
            self.logger.warn(item)

        critical_items = [
            {'location': {'country': 'Canada', 'Province': 'Quebec Ontario', 'city': 'Quebec City'}},
            {'location': {'country': 'Canada', 'Province': 'Quebec Ontario', 'city': 'Toronto'}},
        ]

        for item in critical_items:
            self.logger.critical(item)

        results = Mongolog.find(query={'msg.location': {'$exists': True}})
        self.assertEqual(11, len(list(results)))

        results = Mongolog.find(level='INFO')
        self.assertEqual(6, len(list(results)))

        # Test limit
        results = Mongolog.find(level='INFO', limit=3)
        self.assertEqual(3, len(list(results)))

        # Test search by name
        results = Mongolog.find(logger="test.embedded")
        self.assertEqual(11, len(list(results)))

        results = Mongolog.find(query={'msg.location.region': {'$exists': True}})
        self.assertEqual(4, len(list(results)))

        results = Mongolog.find(query={'msg.location.region': {'$exists': True}}, level='ERROR')
        self.assertEqual(0, len(list(results)))

        results = Mongolog.find(query={'msg.location.region': {'$exists': True}}, level='WARNING')
        self.assertEqual(2, len(list(results)))

        results = Mongolog.find(level='CRITICAL')
        self.assertEqual(2, len(list(results)))

        results = Mongolog.find(level='CRITICAL', query={'msg.location.city': 'Quebec City'})
        self.assertEqual(1, len(list(results)))

        # Test uuid path.
        results = Mongolog.find(uuid='no-results')
        self.assertEqual(0, len(list(results)))

        # Now test projection
        query = {'msg.location.city': 'Quebec City'}
        level = 'CRITICAL'
        results = Mongolog.find(level=level, query=query, project={'msg': 1})
        results = list(results)
        self.assertEqual(set(['_id', 'msg']), set(results[0].keys()))

        results = Mongolog.find(level=level, query=query, project={'msg': 1, '_id': 0})
        results = list(results)
        self.assertEqual(set(['msg']), set(results[0].keys()))

        results = Mongolog.find(level=level, query=query, project={'msg': 1, '_id': 0, 'level': 1})
        results = list(results)
        self.assertEqual(set(['msg', 'level']), set(results[0].keys()))


class FlakeTests(unittest.TestCase):
    # flake8 --ignore=E402,F82
    def test_flak8(self):
        cmd = 'flake8 --ignore=E402,F82'
        logger = logging.getLogger("console")
        logger.debug(self)
        try:
            subprocess.check_output([cmd], shell=True, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError:
            logger.critical("----------------- Please Fix flake8 errors -------------------")
            os.system(cmd)
            logger.critical("--------------------------------------------------------------")
            raise



# -*- coding: utf-8 -*-
"""
Management command to handle deletion of previous mongolog records.

Usage Examples:

# Delete all records alder than 54 day's AFTER backing up the current collection 'bulkupload'
 ./manage.py ml_purge -d 54 -b -c bulkupload

 # Delete all entries
./manage.py ml_purge -p -c bulkupload
"""
from __future__ import print_function
import sys
import logging
from datetime import timedelta
import subprocess
from pymongo import MongoClient

from mongolog.handlers import get_mongolog_handler

from django.utils import timezone
from django.core.management.base import BaseCommand

console = logging.getLogger('mongolog-int')


class Command(BaseCommand):

    def add_arguments(self, parser):
        parser.add_argument(
            '-p', '--purge', default=False, action='store_true', dest='purge',
            help='Remove all old results',
        )
        parser.add_argument(
            '-d', '--delete', default=14, type=int, action='store', dest='delete',
            help='Delete documents more than -d={n} days old',
        )
        parser.add_argument(
            '-f', '--force', default=False, action='store_true', dest='force',
            help='Do not prompt before removing documents',
        )
        parser.add_argument(
            '-b', '--backup', default=False, action='store_true', dest='backup',
            help='Backup collection before deleting',
        )
        parser.add_argument(
            '-l', '--logger', default='mongolog', type=str, action='store', dest='logger',
            help='Which mongolog logger to use.  The collection defined in the log handler will be used.',
        )

    def __init__(self, *args, **kwargs):
        super(Command, self).__init__(*args, **kwargs)

    def confirm(self, **options):
        if not options['force']:
            while 1:
                console.warn("Would you like to proceed?  Y/N")
                ans = input().strip().lower()

                if ans not in ['y', 'yes', 'n', 'no']:
                    continue
                elif ans[0] == 'n':
                    console.info("You chose not to continue.  Bye!")
                    sys.exit(1)
                elif ans[0] == 'y':
                    break

        return True

    def purge(self, **options):
        """ Purge all records from the collection """
        console.warn("You are about to delete all mongolog documents!!!")

        if self.confirm(**options):
            total = self.collection.find({}).count()
            self.collection.delete_many({})
            console.warn("Total docs to remove: %s", total)

    def delete(self, **options):
        """ Delete all records older than --delete={n} """
        days = options['delete']
        console.warn("Looking for documents older than %s day's", days)

        query = {
            'created': {
                '$lte': timezone.now() - timedelta(days=days)
            }
        }

        total = self.collection.find(query).count()
        console.warn("Total docs to remove: %s", total)

        if self.confirm(**options):
            self.collection.delete_many(query)
            console.info("Total docs removed: %s", total)

    def backup(self):
        """ Backup the collection before deleting """
        console.info("Backing up your documents...")
        cmd = 'mongodump --db mongolog'
        subprocess.check_call([cmd], shell=True)

    def handle(self, *args, **options):
        """ Main processing handle """
        handler = get_mongolog_handler(logger_name=options['logger'])
        client = MongoClient(handler.connection)
        db = client.mongolog
        self.collection = getattr(db, handler.collection)

        if options['backup']:
            self.backup()

        if options['purge']:
            self.purge(**options)
        elif isinstance(options['delete'], int):
            self.delete(**options)
